from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, Tuple, Union

from modelon.impact.client import exceptions
from modelon.impact.client.entities.asserts import assert_successful_operation
from modelon.impact.client.entities.custom_artifact import CustomArtifact
from modelon.impact.client.entities.custom_function import CustomFunction
from modelon.impact.client.entities.external_result import ExternalResult
from modelon.impact.client.entities.interfaces.case import CaseReference
from modelon.impact.client.entities.log import Log
from modelon.impact.client.entities.model import (
    Model,
    SimpleModelicaExperimentDefinition,
    to_domain_parameter_value,
)
from modelon.impact.client.entities.model_executable import ModelExecutable
from modelon.impact.client.entities.result import Result
from modelon.impact.client.entities.status import CaseStatus
from modelon.impact.client.experiment_definition.modifiers import Enumeration
from modelon.impact.client.operations.base import BaseOperation
from modelon.impact.client.operations.case import CaseOperation
from modelon.impact.client.operations.case_result import CaseResultImportOperation
from modelon.impact.client.operations.custom_artifact import (
    CustomArtifactImportOperation,
)
from modelon.impact.client.sal.experiment import ResultFormat

if TYPE_CHECKING:
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)


def _assert_case_is_complete(
    status: CaseStatus, operation_name: str = "Operation"
) -> None:
    if status == CaseStatus.NOT_STARTED:
        raise exceptions.OperationNotCompleteError.for_operation(operation_name, status)
    elif status == CaseStatus.CANCELLED:
        raise exceptions.OperationFailureError.for_operation(operation_name)


def _datetime_from_unix_time(unix_time: Optional[int]) -> Optional[datetime]:
    return datetime.fromtimestamp(unix_time / 1e3) if unix_time else None


class CaseRunInfo:
    """Class containing Case run info."""

    def __init__(
        self,
        status: CaseStatus,
        consistent: bool,
        datetime_started: Optional[datetime],
        datetime_finished: Optional[datetime],
    ):
        self._status = status
        self._consistent = consistent
        self._datetime_started = datetime_started
        self._datetime_finished = datetime_finished

    @property
    def status(self) -> CaseStatus:
        """Status info for a Case.

        .. Note:: For tracing status changes of a running case execution,
            the CaseOperation status property must be used and not Case
            entities run info(CaseRunInfo) status property. See example
            for usage.

        Example::

            # To get the case entity run info status
            status = case.run_info.status

            # To get the status updates of a running case execution
            case_ops = case.execute()
            status = case_ops.status

        """
        return self._status

    @property
    def consistent(self) -> bool:
        """True if the case has not been synced since it was executed, false
        otherwise."""
        return self._consistent

    @property
    def started(self) -> Optional[datetime]:
        """Case execution start time.

        Returns None if case execution hasn't started.

        """
        return self._datetime_started

    @property
    def finished(self) -> Optional[datetime]:
        """Case execution finish time.

        Returns None if case execution hasn't finished.

        """
        return self._datetime_finished


class CaseAnalysis:
    """Class containing Case analysis configuration."""

    __slots__ = ["_analysis"]

    def __init__(self, analysis: Dict[str, Any]):
        self._analysis = analysis

    @property
    def analysis_function(self) -> str:
        """Custom function name.

        Returns:
            The name of the custom function.

        Example::

            analysis_function = case.input.analysis.analysis_function

        """
        return self._analysis["type"]

    @property
    def parameters(self) -> Dict[str, Any]:
        """Get or set parameters for the custom function. Parameters are a dictionary
        containing the custom function parameters.

        Example::

            # Set the custom function parameters
            case.input.analysis.parameters = {'start_time': 0, 'final_time': 90}


            # Get the custom function parameters
            parameters = case.input.analysis.parameters

        """
        parameters = self._analysis["parameters"]
        return {param["name"]: param["value"] for param in parameters}

    @parameters.setter
    def parameters(self, parameters: Dict[str, Any]) -> None:
        self._analysis["parameters"] = [
            {"name": param, "value": parameters[param]} for param in parameters
        ]

    @property
    def simulation_options(self) -> Dict[str, Any]:
        """Get or set the simulation options. Options are key-value pairs of simulation
        options.

        Example::

            # Set the custom function parameters
            case.input.analysis.simulation_options = {'ncp': 600}

            # Get the custom function parameters
            simulation_options = case.input.analysis.simulation_options

        """
        return self._analysis["simulationOptions"]

    @simulation_options.setter
    def simulation_options(self, simulation_options: Dict[str, Any]) -> None:
        self._analysis["simulationOptions"] = simulation_options

    @property
    def solver_options(self) -> Dict[str, Any]:
        """Get or set the simulation options. Options are key-value pairs of solver
        options.

        Example::

            # Set the solver options
            case.input.analysis.solver_options = {'atol': 1e-9}


            # Get the solver options
            solver_options = case.input.analysis.solver_options

        """
        return self._analysis["solverOptions"]

    @solver_options.setter
    def solver_options(self, solver_options: Dict[str, Any]) -> None:
        self._analysis["solverOptions"] = solver_options

    @property
    def simulation_log_level(self) -> str:
        """Get or set the simulation log level. Supported options are- 'WARNING',
        'ERROR', 'DEBUG', 'INFO' and 'VERBOSE'.

        Example::

            # Set the solver options
            case.input.analysis.simulation_log_level = "WARNING"

            # Get the solver options
            simulation_log_level = case.input.analysis.simulation_log_level

        """
        return self._analysis["simulationLogLevel"]

    @simulation_log_level.setter
    def simulation_log_level(self, simulation_log_level: str) -> None:
        self._analysis["simulationLogLevel"] = simulation_log_level


class CaseMeta:
    """Class containing Case meta."""

    __slots__ = ["_data"]

    def __init__(self, data: Dict[str, Any]):
        self._data = data

    @property
    def label(self) -> str:
        """Get or set the label for the case.

        Example::

            # Set the case label
            case.meta.label = 'Cruise condition'

            # Get the case label
            label = case.meta.label

        """
        return self._data["label"]

    @label.setter
    def label(self, label: str) -> None:
        self._data["label"] = label

    @property
    def orchestrator(self) -> bool:
        """Returns True, if this is an Orchestrator case which creates and runs other
        cases in this experiment..

        Example::

            orchestrator = case.meta.orchestrator

        """
        return self._data.get("orchestrator", False)


class CaseInput:
    """Class containing Case input."""

    __slots__ = ["_data"]

    def __init__(self, data: Dict[str, Any]):
        self._data = data

    @property
    def analysis(self) -> CaseAnalysis:
        return CaseAnalysis(self._data["analysis"])

    @property
    def parametrization(self) -> Dict[str, Any]:
        """Get or set the parametrization of the case. Parameterization is defined as a
        dict of key value pairs where key is variable name and value is the value to use
        for that variable.

        Example::

            # Set the case parametrization
            case.input.parametrization = {'PI.k': 120}


            # Get the case parametrization
            parametrization = case.input.parametrization

        """
        parametrization = self._data["parametrization"]
        return {
            param["name"]: to_domain_parameter_value(param) for param in parametrization
        }

    @parametrization.setter
    def parametrization(self, parametrization: Dict[str, Any]) -> None:
        converted_params = []
        for name, value in parametrization.items():
            if isinstance(value, Enumeration):
                converted_params.append(value.to_dict(name))
            else:
                converted_params.append({"name": name, "value": value})

        self._data["parametrization"] = converted_params

    @property
    def fmu_id(self) -> str:
        """Reference ID to the compiled model used running the case."""
        return self._data["fmuId"]

    @property
    def structural_parametrization(self) -> Dict[str, Any]:
        """Structural parametrization of the case, a list of key value pairs where key
        is variable name and value is the value to use for that variable.

        These are values that cannot be applied to the FMU/Model after compilation.

        """
        parametrization = self._data["structuralParametrization"]
        return {
            param["name"]: to_domain_parameter_value(param) for param in parametrization
        }

    @property
    def fmu_base_parametrization(self) -> Dict[str, Any]:
        """This is some base parametrization that must be applied to the FMU for it to
        be valid running this case.

        It often comes as a result from of caching to reuse the FMU.

        """
        parametrization = self._data["fmuBaseParametrization"]
        return {
            param["name"]: to_domain_parameter_value(param) for param in parametrization
        }


class Case(CaseReference):
    """Class containing Case functionalities."""

    __slots__ = ["_case_id", "_workspace_id", "_exp_id", "_sal", "_info"]

    def __init__(
        self,
        case_id: str,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Dict[str, Any],
    ):
        self._case_id = case_id
        self._workspace_id = workspace_id
        self._exp_id = exp_id
        self._sal = service
        self._info = info
        super().__init__(case_id, workspace_id, exp_id, service, info)

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Case) and obj._case_id == self._case_id

    def __repr__(self) -> str:
        return f"Case with id '{self._case_id}'"

    def _get_info(self, cached: bool = True) -> Dict[str, Any]:
        if not cached or self._info is None:
            self._info = self._sal.experiment.case_get(
                self._workspace_id, self._exp_id, self.id
            )

        return self._info

    @property
    def info(self) -> Dict[str, Any]:
        """Deprecated, use 'run_info' attribute."""
        logger.warning("This attribute is deprectated, use 'run_info' instead")
        return self._get_info(cached=False)

    @property
    def run_info(self) -> CaseRunInfo:
        """Case run information."""
        run_info = self._get_info(cached=False)["run_info"]
        started = _datetime_from_unix_time(run_info.get("datetime_started"))
        finished = _datetime_from_unix_time(run_info.get("datetime_finished"))
        return CaseRunInfo(
            CaseStatus(run_info["status"]), run_info["consistent"], started, finished
        )

    @property
    def input(self) -> CaseInput:
        """Case input attributes.

        Example::

            case.input.analysis.parameters = {'start_time': 0, 'final_time': 90}
            case.input.analysis.simulation_options = {'ncp': 600}
            case.input.analysis.solver_options = {'atol': 1e-8}
            case.input.parametrization = {'PI.k': 120}
            case.sync()

            help(case.input.analysis) # See help for attribute
            dir(case.input) # See nested attributes

        """
        return CaseInput(self._info["input"])

    @property
    def meta(self) -> CaseMeta:
        """Case meta attributes.

        Example::

            case.meta.label = 'Cruise condition'
            case.sync()

            help(case.meta) # See help for attribute
            dir(case.input) # See nested attributes

        """
        return CaseMeta(self._info["meta"])

    @property
    def initialize_from_case(self) -> Optional[Case]:
        """Get(if any) or set the case to initialize from.

        Example::

            # Set the case to  initialize from
            # Fetching the successful case
            case_1 = experiment.get_case('case_1')

            # Fetching the failed case
            case_2 = experiment.get_case('case_2')

            # Initializing from successful case
            case_2.initialize_from_case = case_1

            # Re-executing the case after initializing
            case_init_successful = case_2.execute().wait()


            # Get the case set to initialize from
            initialized_from_case = case.initialize_from_case

        """
        init_from_dict = self._info["input"].get("initializeFromCase")
        if init_from_dict is None:
            return None

        experiment_id = init_from_dict.get("experimentId")
        case_id = init_from_dict.get("caseId")

        case_data = self._sal.experiment.case_get(
            self._workspace_id, experiment_id, case_id
        )
        return Case(
            case_data["id"], self._workspace_id, experiment_id, self._sal, case_data
        )

    @initialize_from_case.setter
    def initialize_from_case(self, case: Case) -> None:
        if not isinstance(case, Case):
            raise TypeError("The value must be an instance of Case")
        self._assert_unique_case_initialization(
            "initializeFromExternalResult", "initialize_from_external_result"
        )
        self._info["input"]["initializeFromCase"] = {
            "experimentId": case.experiment_id,
            "caseId": case.id,
        }

    @property
    def initialize_from_external_result(self) -> Optional[ExternalResult]:
        """Get(if any) or set the external result file to initialize from.

        Example::

            # Set the external result file to  initialize from
            result = workspace.upload_result(path_to_result="<path_to_result>/
            result.mat", label = "result_to_init", description= "Converged
            result file").wait()

            # Initializing from external result
            case_2.initialize_from_external_result = result

            # Re-executing the case after initializing
            case_init_successful = case_2.execute().wait()


            # Get the external result file to initialize from
            initialize_from_external_result = case.initialize_from_external_result

        """
        init_from_dict = self._info["input"].get("initializeFromExternalResult")

        if init_from_dict is None:
            return None

        result_id = init_from_dict.get("uploadId")

        return ExternalResult(result_id, self._sal)

    @initialize_from_external_result.setter
    def initialize_from_external_result(self, result: ExternalResult) -> None:
        if not isinstance(result, ExternalResult):
            raise TypeError("The value must be an instance of ExternalResult")
        self._assert_unique_case_initialization(
            "initializeFromCase", "initialize_from_case"
        )
        self._info["input"]["initializeFromExternalResult"] = {"uploadId": result.id}

    def is_successful(self) -> bool:
        """Returns True if a case has completed successfully.

        Returns:
            True, if the case has executed successfully, False,
            if the case has failed execution.

        Example::

            case.is_successful()

        """
        return self.run_info.status == CaseStatus.SUCCESSFUL

    def get_log(self) -> Log:
        """Returns the log class object for a finished case.

        Returns:
            The case execution log class object.

        Example::

            log = case.get_log()
            log.show()

        """
        return Log(
            self._sal.experiment.case_get_log(
                self._workspace_id, self._exp_id, self._case_id
            )
        )

    def get_result(self, format: str = "mat") -> Tuple[Union[bytes, Text], str]:
        """Returns the result stream and the file name for a finished case.

        Args:
            format: The file format to download the result in. The only possible values
                are 'mat' and 'csv'. Default: 'mat'

        Returns:
            A tuple containing, respectively, the result byte stream.
            and the filename for the result. This name could be used to
            write the result stream.

        Raises:

            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.
            TypeError if the variable is not a list object.
            ValueError if trajectory variable is not present in the result.

        Example::

            result, file_name = case.get_result(format = 'csv')
            with open(file_name, "w") as f:
                f.write(result)

        """
        assert_successful_operation(self.is_successful(), self._case_id)
        result_format = ResultFormat(format)
        result, file_name = self._sal.experiment.case_result_get(
            self._workspace_id, self._exp_id, self._case_id, result_format
        )
        return result, file_name

    def get_trajectories(self) -> Result:
        """Returns result(Mapping) object containing the result trajectories.

        Returns:
            A Result object (mapping) that allows requesting trajectory data given
            a variable name.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process was cancelled.

        Example::

            result = case.get_trajectories()
            result_variables = result.keys()
            height = result['h']
            time = res['time']

        """
        _assert_case_is_complete(self.run_info.status, "Simulation")
        return Result(
            self._sal.experiment.case_result_variables_get(
                self._workspace_id, self._exp_id, self._case_id
            ),
            self._case_id,
            self._workspace_id,
            self._exp_id,
            self._sal,
        )

    def get_variables(self) -> List[str]:
        """Returns a list of variables available in the result.

        Returns:
            An list of result variables.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.

        Example::

            case.get_variables()

        """
        _assert_case_is_complete(self.run_info.status, "Simulation")
        return self._sal.experiment.case_result_variables_get(
            self._workspace_id, self._exp_id, self._case_id
        )

    def get_artifact(
        self, artifact_id: str, download_as: Optional[str] = None
    ) -> CustomArtifact:
        """Returns a CustomArtifact class for a finished case.

        Returns:
            The CustomArtifact class object.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.

        Example::

            custom_artifact = case.get_artifact(artifact_id)

        """
        assert_successful_operation(self.is_successful(), self._case_id)
        return CustomArtifact(
            self._workspace_id,
            self.experiment_id,
            self._case_id,
            artifact_id,
            download_as
            if download_as
            else self._get_artifact_download_name(artifact_id),
            self._sal.experiment,
        )

    def _get_artifact_download_name(self, artifact_id: str) -> str:
        resp = self._sal.experiment.case_artifacts_meta_get(
            self._workspace_id, self._exp_id, self._case_id
        )
        meta = next(
            (meta for meta in resp["data"]["items"] if meta["id"] == artifact_id),
            None,
        )
        if not meta:
            raise exceptions.NoSuchCustomArtifactError(
                f"No custom artifact found with ID: {artifact_id}."
            )
        return meta["downloadAs"]

    def get_artifacts(self) -> List[CustomArtifact]:
        """Returns a list of CustomArtifact classes for a finished case.

        Returns:
            A list of CustomArtifact class objects.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.

        Example::

            custom_artifacts = case.get_artifacts()

        """
        assert_successful_operation(self.is_successful(), self._case_id)
        resp = self._sal.experiment.case_artifacts_meta_get(
            self._workspace_id, self._exp_id, self._case_id
        )
        return [
            CustomArtifact(
                self._workspace_id,
                self.experiment_id,
                self._case_id,
                meta["id"],
                meta["downloadAs"],
                self._sal.experiment,
            )
            for meta in resp["data"]["items"]
        ]

    def get_fmu(
        self,
    ) -> ModelExecutable:
        """Returns the ModelExecutable class object simulated for the case.

        Returns:
            ModelExecutable class object.

        Example::

            case = experiment.get_case('case_1')
            fmu = case.get_fmu()
            fmus = set(case.get_fmu() for case in exp.get_cases())

        """
        fmu_id = self.input.fmu_id
        if not fmu_id:
            custom_function = self._get_custom_function()
            for param in custom_function.parameter_values.values():
                if isinstance(param, CaseReference):
                    case = self.from_reference(param)
                    fmu_id = case.input.fmu_id
        return ModelExecutable(self._workspace_id, fmu_id, self._sal)

    def sync(self) -> None:
        """Sync case state against server, pushing any changes that has been done to the
        object client side.

        Example::
            case.input.parametrization = {'PI.k': 120}
            case.sync()

        """
        self._info = self._sal.experiment.case_put(
            self._workspace_id, self._exp_id, self._case_id, self._info
        )

    def execute(self, sync_case_changes: bool = True) -> CaseOperation:
        """Executes a case. Returns an CaseOperation class object.

        Args:
            sync_case_changes: Boolean specifying if to sync case changes
            against the server before executing the case. Default is True.


        Returns:
            An CaseOperation class object.

        Example::

            case = experiment.get_case('case_1')
            case.input.parametrization = {'PI.k': 120}
            case.sync()
            case_ops = case.execute()
            case_ops.cancel()
            case_ops.status
            case_ops.wait()

        """
        if sync_case_changes:
            self.sync()

        return CaseOperation[Case](
            self._workspace_id,
            self._sal.experiment.experiment_execute(
                self._workspace_id, self._exp_id, [self._case_id]
            ),
            self._case_id,
            self._sal,
            Case.from_operation,
        )

    def _get_custom_function(self) -> CustomFunction:
        custom_function_meta = self._sal.custom_function.custom_function_get(
            self._workspace_id, self.input.analysis.analysis_function
        )
        custom_function_params = self.input.analysis.parameters
        return CustomFunction(
            self._workspace_id,
            custom_function_meta["name"],
            custom_function_meta["parameters"],
            self._sal,
        ).with_parameters(**custom_function_params)

    def get_definition(self) -> SimpleModelicaExperimentDefinition:
        """Get an experiment definition that can be used to reproduce this case result.

        Returns:
            An instance of SimpleModelicaExperimentDefinition class.

        Example::

            definition = case.get_definition()

        """
        custom_function = self._get_custom_function()
        fmu = self.get_fmu()
        model = Model(
            fmu.input.class_name,
            workspace_id=self._workspace_id,
            project_id="",
            service=self._sal,
        )
        definition = SimpleModelicaExperimentDefinition(
            model=model,
            custom_function=custom_function,
            compiler_options=fmu.input.compiler_options,
            fmi_target=fmu.input.fmi_target,
            fmi_version=fmu.input.fmi_version,
            platform=fmu.input.platform,
            compiler_log_level=fmu.input.compiler_log_level,
            runtime_options=fmu.input.runtime_options,
            solver_options=self.input.analysis.solver_options,
            simulation_options=self.input.analysis.simulation_options,
            simulation_log_level=self.input.analysis.simulation_log_level,
            initialize_from=self.initialize_from_case
            or self.initialize_from_external_result,
        )
        modifiers = {
            **self.input.structural_parametrization,
            **self.input.parametrization,
        }
        definition = definition.with_modifiers(modifiers=modifiers)
        return definition

    def _assert_unique_case_initialization(
        self, unsupported_init: str, unsupported_method_name: str
    ) -> None:
        if self._info["input"][unsupported_init]:
            raise ValueError(
                "A case cannot use both 'initialize_from_case' and "
                "'initialize_from_external_result' to specify what to initialize from! "
                f"To resolve this, set the '{unsupported_method_name}' attribute "
                "to None and re-try."
            )

    @classmethod
    def from_operation(cls, operation: BaseOperation[Case], **kwargs: Any) -> Case:
        assert isinstance(operation, CaseOperation)
        return cls(**kwargs, service=operation._sal)

    @classmethod
    def from_reference(cls, reference: CaseReference) -> Case:
        return cls(
            case_id=reference._case_id,
            workspace_id=reference._workspace_id,
            exp_id=reference._exp_id,
            service=reference._sal,
            info=reference._info,
        )

    def import_custom_artifact(
        self,
        path_to_artifact: str,
        artifact_id: Optional[str] = None,
        overwrite: bool = False,
    ) -> CustomArtifactImportOperation:
        """Upload custom artifact to a case.

        Args:

            path_to_artifact: The path for the artifact to be imported.
            artifact_id: ID of the artifact to be imported.
            overwrite: Overwrite, if any already existing artifact exists
                        with the same ID. Default: False.


        Example::

            case.import_custom_artifact('/home/test.csv', artifact_id).wait()

        """
        resp = self._sal.experiment.custom_artifact_upload(
            path_to_artifact,
            self._workspace_id,
            self._exp_id,
            self._case_id,
            artifact_id,
            overwrite,
        )
        return CustomArtifactImportOperation[CustomArtifact](
            resp["data"]["location"],
            self._workspace_id,
            self._exp_id,
            self._case_id,
            self._sal,
            CustomArtifact.from_operation,
        )

    def import_result(
        self,
        path_to_result: str,
        overwrite: bool = False,
    ) -> CaseResultImportOperation:
        """Upload result to a case.

        Args:

            path_to_result: The path for the result to be imported. Only mat or csv
                result file format is supported for import.
            overwrite: Overwrite, if a result already exists
                for the case. Default: False.


        Example::

            case.import_result('/home/test.csv').wait()

        """
        resp = self._sal.experiment.case_result_upload(
            path_to_result,
            self._workspace_id,
            self._exp_id,
            self._case_id,
            overwrite,
        )
        return CaseResultImportOperation[Result](
            resp["data"]["location"],
            self._workspace_id,
            self._exp_id,
            self._case_id,
            self._sal,
            Result.from_operation,
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.entities.interfaces.case import CaseInterface
from modelon.impact.client.entities.interfaces.experiment import ExperimentInterface
from modelon.impact.client.entities.interfaces.external_result import (
    ExternalResultInterface,
)
from modelon.impact.client.experiment_definition.asserts import (
    assert_valid_args,
    assert_valid_case_modifiers,
    assert_valid_extensions,
    validate_initialize_from,
)
from modelon.impact.client.experiment_definition.expansion import (
    ExpansionAlgorithm,
    FullFactorial,
)
from modelon.impact.client.experiment_definition.extension import (
    SimpleExperimentExtension,
)
from modelon.impact.client.experiment_definition.interfaces.definition import (
    BaseExperimentDefinition,
)
from modelon.impact.client.experiment_definition.modifiers import (
    Modifier,
    ensure_as_modifier,
    modifiers_to_dict,
)
from modelon.impact.client.experiment_definition.util import (
    case_to_identifier_dict,
    custom_function_parameters_to_dict,
    get_options,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case
    from modelon.impact.client.entities.custom_function import CustomFunction
    from modelon.impact.client.entities.experiment import Experiment
    from modelon.impact.client.entities.external_result import ExternalResult
    from modelon.impact.client.entities.model import Model
    from modelon.impact.client.options import (
        CompilerOptions,
        RuntimeOptions,
        SimulationOptions,
        SolverOptions,
    )

    CaseOrExperimentOrExternalResult = Union[Case, Experiment, ExternalResult]
    RuntimeOptionsOrDict = Union[RuntimeOptions, Dict[str, Any]]
    SimulationOptionsOrDict = Union[SimulationOptions, Dict[str, Any]]
    SolverOptionsOrDict = Union[SolverOptions, Dict[str, Any]]
    CompilerOptionsOrDict = Union[CompilerOptions, Dict[str, Any]]

logger = logging.getLogger(__name__)


class SimpleModelicaExperimentDefinition(BaseExperimentDefinition):
    """A simple experiment definition class for defining experiments.

    Args:
        model: The Model class object.
        custom_function:bThe custom function to use for this experiment.
        compiler_options:bThe compiler options to use for this experiment.
            By default the options is set to None, which means the default
            options for the custom_function input is used.
        fmi_target: Compiler target. Possible values are 'me' and 'cs'. Default: 'me'.
        fmi_version: The FMI version. Valid options are '1.0' and '2.0'. Default: '2.0'.
        platform: Platform for FMU binary.The OS running the Impact server must match
            the environment that runs the compiled FMU. This is necessary as the
            binaries packaged with the FMU are based on the platform generating
            the FMU. For example, if the Impact server is running Linux the binary
            in the downloaded FMU is compiled for Linux. The downloaded FMU can
            then not be simulated on Windows. Default: 'auto'. Supported options are:-

            - 'auto': platform is selected automatically
            - 'linux64': generate a 32 bit FMU
            - 'win64': generate a 64 bit FMU
        compiler_log_level: The logging for the compiler. Possible values are "error",
            "warning", "info", "verbose" and "debug". Default: 'warning'.
        runtime_options: The runtime options to use for this experiment. By default the
            options is set to None, which means the default options for the
            custom_function input is used.
        solver_options: The solver options to use for this experiment. By default the
            options is set to None, which means the default options for the
            custom_function input is used.
        simulation_options: The simulation options to use for this experiment. By
            default the options is set to None, which means the default options for the
            custom_function input is used.
        simulation_log_level: Simulation log level for this experiment.
            Default: 'WARNING'.
        initialize_from: Optional entity to initialize from. An instance of
            Case or Experiment or ExternalResult. Default: None

    Example::

        model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
        simulation_options = custom_function.get_simulation_options()
        .with_values(ncp=500)
        solver_options = {'atol':1e-8}
        simulate_def = model.new_experiment_definition(
            custom_function,
            solver_options=solver_options,
            simulation_options=simulation_options
        )
        simulate_def.to_dict()

    """

    def __init__(
        self,
        model: Model,
        custom_function: CustomFunction,
        *,
        compiler_options: Optional[CompilerOptionsOrDict] = None,
        fmi_target: str = "me",
        fmi_version: str = "2.0",
        platform: str = "auto",
        compiler_log_level: str = "warning",
        runtime_options: Optional[RuntimeOptionsOrDict] = None,
        solver_options: Optional[SolverOptionsOrDict] = None,
        simulation_options: Optional[SimulationOptionsOrDict] = None,
        simulation_log_level: str = "WARNING",
        initialize_from: Optional[CaseOrExperimentOrExternalResult] = None,
    ):
        assert_valid_args(
            model=model,
            custom_function=custom_function,
            compiler_options=compiler_options,
            runtime_options=runtime_options,
            solver_options=solver_options,
            simulation_options=simulation_options,
        )
        self._model = model
        self._custom_function = custom_function
        self._compiler_options = get_options(
            custom_function.get_compiler_options, compiler_options
        )
        self._runtime_options = get_options(
            custom_function.get_runtime_options, runtime_options
        )
        self._fmi_target = fmi_target
        self._fmi_version = fmi_version
        self._platform = platform
        self._compiler_log_level = compiler_log_level
        self._solver_options = get_options(
            custom_function.get_solver_options, solver_options
        )
        self._simulation_options = get_options(
            custom_function.get_simulation_options, simulation_options
        )
        self._simulation_log_level = simulation_log_level
        self._initialize_from = initialize_from
        self._variable_modifiers: Dict[str, Modifier] = {}
        self._extensions: List[SimpleExperimentExtension] = []
        self._expansion: ExpansionAlgorithm = FullFactorial()

    @property
    def modifiers(self) -> Dict[str, Any]:
        """Returns the variable modifiers dict."""
        return self._variable_modifiers

    @property
    def model(self) -> Model:
        """Returns the Model class."""
        return self._model

    @property
    def fmi_target(self) -> str:
        """Returns the FMI target."""
        return self._fmi_target

    @property
    def fmi_version(self) -> str:
        """Returns the FMI version."""
        return self._fmi_version

    @property
    def platform(self) -> str:
        """Returns the platform to compile FMU for."""
        return self._platform

    @property
    def compiler_log_level(self) -> str:
        """Returns the compiler log level."""
        return self._compiler_log_level

    @property
    def compiler_options(self) -> Dict[str, Any]:
        """Returns the compiler options as a dict."""
        return self._compiler_options

    @property
    def runtime_options(self) -> Dict[str, Any]:
        """Returns the runtime options as a dict."""
        return self._runtime_options

    @property
    def simulation_options(self) -> Dict[str, Any]:
        """Returns the simulation options as a dict."""
        return self._simulation_options

    @property
    def solver_options(self) -> Dict[str, Any]:
        """Returns the solver options as a dict."""
        return self._solver_options

    @property
    def simulation_log_level(self) -> str:
        """Returns the simulation log level."""
        return self._simulation_log_level

    @property
    def custom_function(self) -> CustomFunction:
        """Returns the custom function class."""
        return self._custom_function

    @property
    def extensions(self) -> List[SimpleExperimentExtension]:
        """Returns the list of experiment extensions."""
        return self._extensions

    @property
    def expansion(self) -> ExpansionAlgorithm:
        """Returns the expansion algorithm class."""
        return self._expansion

    @property
    def initialize_from(self) -> Optional[CaseOrExperimentOrExternalResult]:
        """Returns the case, experiment or result the experiment definition was
        initialized from."""
        return self._initialize_from

    def validate(self) -> None:
        raise NotImplementedError(
            "Validation is not supported for SimpleModelicaExperimentDefinition class"
        )

    def with_modifiers(
        self,
        modifiers: Optional[Dict[str, Any]] = None,
    ) -> SimpleModelicaExperimentDefinition:
        """Sets the modifiers parameters for an experiment.

        Args:
            modifiers: A dictionary of variable modifiers. Could be used if
                modifiers keys conflict with python identifiers or keywords.
                Default: None.

        Example::

            from modelon.impact.client import Range, Choices

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            experiment_definition = model.new_experiment_definition(
                custom_function).with_modifiers({'inertia1.J': Choices(0.1, 0.9),
                'inertia2.J': Range(0.1, 0.5, 3)})

        """
        modifiers = {} if modifiers is None else modifiers
        new = SimpleModelicaExperimentDefinition(
            model=self._model,
            custom_function=self._custom_function,
            compiler_options=self._compiler_options,
            fmi_target=self._fmi_target,
            fmi_version=self._fmi_version,
            platform=self._platform,
            compiler_log_level=self._compiler_log_level,
            runtime_options=self._runtime_options,
            solver_options=self._solver_options,
            simulation_options=self._simulation_options,
            simulation_log_level=self._simulation_log_level,
            initialize_from=self._initialize_from,
        )
        new._extensions = self._extensions
        new._expansion = self._expansion
        new._variable_modifiers = self._variable_modifiers

        for name, value in modifiers.items():
            new._variable_modifiers[name] = ensure_as_modifier(value)
        return new

    def with_expansion(
        self, expansion: Optional[ExpansionAlgorithm] = None
    ) -> SimpleModelicaExperimentDefinition:
        """Sets the expansion algorithm for an experiment.

        Args:
            expansion: An expansion algorithm. Available algorithms are LatinHypercube,
                Sobol and FullFactorial.
                Default: FullFactorial.

        Example::

            from modelon.impact.client import Sobol, Beta, Normal

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            experiment_definition = model.new_experiment_definition(
                custom_function).with_modifiers({'inertia1.J': Beta(0.1, 0.9),
                'inertia2.J': Normal(0.1, 0.5)}).with_expansion(Sobol(5))

        """
        if not isinstance(expansion, ExpansionAlgorithm):
            raise TypeError(
                f"The expansion argument is of type '{type(expansion)}' "
                "which is not a subtype of 'ExpansionAlgorithm'!"
            )
        new_expansion = expansion or FullFactorial()
        new = SimpleModelicaExperimentDefinition(
            model=self._model,
            custom_function=self._custom_function,
            compiler_options=self._compiler_options,
            fmi_target=self._fmi_target,
            fmi_version=self._fmi_version,
            platform=self._platform,
            compiler_log_level=self._compiler_log_level,
            runtime_options=self._runtime_options,
            solver_options=self._solver_options,
            simulation_options=self._simulation_options,
            simulation_log_level=self._simulation_log_level,
            initialize_from=self._initialize_from,
        )
        new._extensions = self._extensions
        new._expansion = new_expansion
        new._variable_modifiers = self._variable_modifiers
        return new

    def with_initialize_from(
        self, entity: Optional[CaseOrExperimentOrExternalResult] = None
    ) -> SimpleModelicaExperimentDefinition:
        """Sets the experiment or case to initialize from for an experiment.

        Args:
            entity: An instance of Case or Experiment or ExternalResult.

        Example::

            experiment = workspace.get_experiment(experiment_id)
            experiment_definition = model.new_experiment_definition(custom_function)
            experiment_definition.with_initialize_from(experiment)

            experiment = workspace.get_experiment(experiment_id)
            case = experiment.get_case('case_1')
            experiment_definition = model.new_experiment_definition(custom_function)
            experiment_definition.with_initialize_from(case)

            result = workspace.upload_result('A.mat').wait()
            experiment_definition = model.new_experiment_definition(custom_function)
            experiment_definition.with_initialize_from(result)

        """
        validate_initialize_from(entity)
        new = SimpleModelicaExperimentDefinition(
            model=self._model,
            custom_function=self._custom_function,
            compiler_options=self._compiler_options,
            fmi_target=self._fmi_target,
            fmi_version=self._fmi_version,
            platform=self._platform,
            compiler_log_level=self._compiler_log_level,
            runtime_options=self._runtime_options,
            solver_options=self._solver_options,
            simulation_options=self._simulation_options,
            simulation_log_level=self._simulation_log_level,
            initialize_from=entity,
        )
        new._expansion = self._expansion
        new._variable_modifiers = self._variable_modifiers
        new._extensions = self._extensions
        return new

    def with_extensions(
        self, experiment_extensions: List[SimpleExperimentExtension]
    ) -> SimpleModelicaExperimentDefinition:
        """Sets up an experiment with multiple experiment extensions.

        Args:
            experiment_extensions:
                A list of experiment extension objects.
                Extension object must an instance of
                SimpleExperimentExtension class.

        Example::

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            experiment_definition = model.new_experiment_definition(custom_function).
            with_extensions(
                [
                    SimpleExperimentExtension().with_modifiers({'PI.k': 20}),
                    SimpleExperimentExtension().with_modifiers({'PI.k': 30}),
                ]
            )

            experiment_definition = fmu.new_experiment_definition(custom_function).
            with_extensions(
                [
                    SimpleExperimentExtension(
                        parameter_modifiers={'start_time': 0.0, 'final_time': 2.0}
                    ).with_modifiers({'PI.k': 20})
                ]
            )

        """

        assert_valid_extensions(experiment_extensions)
        exp_ext = []
        for extension in experiment_extensions:
            exp_ext.append(extension)

        new = SimpleModelicaExperimentDefinition(
            model=self._model,
            custom_function=self._custom_function,
            compiler_options=self._compiler_options,
            fmi_target=self._fmi_target,
            fmi_version=self._fmi_version,
            platform=self._platform,
            compiler_log_level=self._compiler_log_level,
            runtime_options=self._runtime_options,
            solver_options=self._solver_options,
            simulation_options=self._simulation_options,
            simulation_log_level=self._simulation_log_level,
            initialize_from=self._initialize_from,
        )
        new._variable_modifiers = self._variable_modifiers
        new._extensions = self._extensions + exp_ext
        new._expansion = self._expansion
        return new

    def with_cases(
        self, cases_modifiers: List[Dict[str, Any]]
    ) -> SimpleModelicaExperimentDefinition:
        """Sets up an experiment with multiple cases with different variable modifiers.

        Args:
            cases_modifiers: A list of variable modifier dictionaries.
                Multiple dictionaries with variable modifiers could to added to create
                multiple cases.

        Example::

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            experiment_definition = model.new_experiment_definition(
                custom_function).with_cases([{'PI.k': 20}, {'PI.k': 30}])

        """
        assert_valid_case_modifiers(cases_modifiers)

        extensions = [
            SimpleExperimentExtension().with_modifiers(modifiers)
            for modifiers in cases_modifiers
        ]
        return self.with_extensions(extensions)

    def to_dict(self) -> Dict[str, Any]:
        """Returns the experiment definition as a dictionary.

        Returns:
            A dictionary containing the experiment definition.

        Example::

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            simulation_options = custom_function.get_simulation_options()
                .with_values(ncp=500)
            solver_options = {'atol':1e-8}
            simulate_def = model.new_experiment_definition(
                custom_function,
                solver_options=solver_options,
                simulation_options=simulation_options
            )
            simulate_def.to_dict()

        """
        custom_function_parameters = custom_function_parameters_to_dict(
            self._custom_function.parameter_values.as_raw_dict()
        )
        variable_modifiers = modifiers_to_dict(self._variable_modifiers)
        exp_dict: Dict[str, Any] = {
            "experiment": {
                "version": 3,
                "base": {
                    "model": {
                        "modelica": {
                            "className": self._model.name,
                            "compilerOptions": self._compiler_options,
                            "runtimeOptions": self._runtime_options,
                            "compilerLogLevel": self._compiler_log_level,
                            "fmiTarget": self._fmi_target,
                            "fmiVersion": self._fmi_version,
                            "platform": self._platform,
                        }
                    },
                    "modifiers": {"variables": variable_modifiers},
                    "analysis": {
                        "type": self._custom_function.name,
                        "parameters": custom_function_parameters,
                        "simulationOptions": self._simulation_options,
                        "solverOptions": self._solver_options,
                        "simulationLogLevel": self._simulation_log_level,
                    },
                    "expansion": {"algorithm": str(self._expansion)},
                },
                "extensions": [ext.to_dict() for ext in self._extensions],
            }
        }
        expansion_parameters = self._expansion.get_parameters_as_dict()
        if expansion_parameters is not None:
            exp_dict["experiment"]["base"]["expansion"][
                "parameters"
            ] = expansion_parameters
        if isinstance(self.initialize_from, ExperimentInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFrom"
            ] = self.initialize_from.id
        elif isinstance(self.initialize_from, CaseInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFromCase"
            ] = case_to_identifier_dict(self.initialize_from)
        elif isinstance(self.initialize_from, ExternalResultInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFromExternalResult"
            ] = self.initialize_from.id
        return exp_dict



from __future__ import annotations

import enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional

from modelon.impact.client.operations.content_import import ContentImportOperation
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation


@enum.unique
class ContentType(enum.Enum):
    """Supported content types in a project."""

    MODELICA = "MODELICA"
    VIEWS = "VIEWS"
    FAVORITES = "FAVORITES"
    CUSTOM_FUNCTIONS = "CUSTOM_FUNCTIONS"
    REFERENCE_RESULTS = "REFERENCE_RESULTS"
    EXPERIMENT_DEFINITIONS = "EXPERIMENT_DEFINITIONS"
    GENERIC = "GENERIC"


class ProjectContent:
    """Content entry in a project."""

    def __init__(self, content: Dict[str, str], project_id: str, service: Service):
        self._content = content
        self._project_id = project_id
        self._sal = service

    def __repr__(self) -> str:
        return f"Project content with id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ProjectContent) and obj.id == self.id

    @property
    def relpath(self) -> Path:
        """Relative path in the project.

        Can be file (e.g., SomeLib.mo) or folder

        """
        return Path(self._content["relpath"])

    @property
    def content_type(self) -> ContentType:
        """Type of content."""
        return ContentType(self._content["contentType"])

    @property
    def id(self) -> str:
        """Content ID."""
        return self._content["id"]

    @property
    def name(self) -> Optional[str]:
        """Modelica library name."""
        return self._content.get("name")

    @property
    def default_disabled(self) -> str:
        return self._content["defaultDisabled"]

    def delete(self) -> None:
        """Deletes a project content.

        Example::

            content.delete()

        """
        self._sal.project.project_content_delete(self._project_id, self.id)

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[ProjectContent], **kwargs: Any
    ) -> ProjectContent:
        assert isinstance(operation, ContentImportOperation)
        return cls(**kwargs, service=operation._sal)



"""Experiment service module."""
from __future__ import annotations

import enum
import json
from typing import Any, Dict, List, Optional, Text, Tuple, Union

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


@enum.unique
class ResultFormat(enum.Enum):
    """Class representing an enumeration for the possible formats available for
    downloading results."""

    MAT = "mat"
    CSV = "csv"

    @classmethod
    def _missing_(cls, value: Any) -> Any:
        if cls not in ["mat", "csv"]:
            raise ValueError(
                "Invalid result format! Allowed formats are 'mat' and 'csv."
            )
        return cls(value)


class ExperimentService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client
        self._case_schema = "application/vnd.impact.cases.v2+json"

    def experiment_execute(
        self, workspace_id: str, exp_id: str, case_ids: Optional[List[str]] = None
    ) -> str:
        body = {"includeCases": {"ids": case_ids}} if case_ids is not None else None
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{exp_id}/execution"
        ).resolve()
        self._http_client.post_json_no_response_body(url, body=body)
        return exp_id

    def experiment_delete(self, workspace_id: str, exp_id: str) -> None:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/experiments/{exp_id}"
        ).resolve()
        self._http_client.delete_json(url)

    def experiment_set_label(self, workspace_id: str, exp_id: str, label: str) -> None:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/experiments/{exp_id}"
        ).resolve()
        return self._http_client.put_json_no_response_body(url, body={"label": label})

    def execute_status(self, workspace_id: str, experiment_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/execution"
        ).resolve()
        return self._http_client.get_json(url)

    def execute_cancel(self, workspace_id: str, experiment_id: str) -> None:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/execution"
        ).resolve()
        return self._http_client.delete_json(url)

    def experiment_result_variables_get(
        self, workspace_id: str, experiment_id: str
    ) -> List[str]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/variables"
        ).resolve()
        return self._http_client.get_json(url)

    def case_result_variables_get(
        self, workspace_id: str, experiment_id: str, case_id: str
    ) -> List[str]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/variables"
        ).resolve()
        return self._http_client.get_json(url)

    def trajectories_get(
        self,
        workspace_id: str,
        experiment_id: str,
        variables: List[str],
        last_point_only: bool,
        format: Optional[str] = None,
    ) -> Any:
        body = {
            "variable_names": variables,
            "filter": {"lastPointOnly": last_point_only},
        }
        headers = {"Accept": format or "application/vnd.impact.trajectories.v1+json"}
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/trajectories"
        ).resolve()
        return self._http_client.post_json(url, body=body, headers=headers)

    def cases_get(self, workspace_id: str, experiment_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases"
        ).resolve()
        return self._http_client.get_json(url, headers={"Accept": self._case_schema})

    def case_get(
        self, workspace_id: str, experiment_id: str, case_id: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}"
        ).resolve()
        return self._http_client.get_json(url, headers={"Accept": self._case_schema})

    def case_put(
        self,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        case_data: Dict[str, Any],
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}"
        ).resolve()
        return self._http_client.put_json(
            url,
            body=case_data,
            headers={"Content-type": self._case_schema, "Accept": self._case_schema},
        )

    def case_get_log(self, workspace_id: str, experiment_id: str, case_id: str) -> str:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/log"
        ).resolve()
        return self._http_client.get_text(url)

    def case_result_get(
        self,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        result_format: ResultFormat = ResultFormat.MAT,
    ) -> Tuple[Union[Text, bytes], str]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/result"
        ).resolve()
        if result_format == ResultFormat.CSV:
            headers = {"Accept": "text/csv"}
            csv_resp = self._http_client.get_csv(url, headers=headers)
            return csv_resp.stream, csv_resp.file_name
        headers = {"Accept": "application/vnd.impact.mat.v1+octet-stream"}
        mat_resp = self._http_client.get_mat(url, headers=headers)
        return mat_resp.stream, mat_resp.file_name

    def case_trajectories_get(
        self,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        variables: List[str],
        last_point_only: bool,
    ) -> List[str]:
        body = {
            "variable_names": variables,
            "filter": {"lastPointOnly": last_point_only},
        }
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/trajectories"
        ).resolve()
        return self._http_client.post_json(url, body=body)

    def case_artifact_get(
        self, workspace_id: str, experiment_id: str, case_id: str, artifact_id: str
    ) -> Tuple[bytes, str]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/custom-artifacts/{artifact_id}"
        ).resolve()
        resp = self._http_client.get_octet_response(url)
        return resp.stream, resp.file_name

    def case_artifacts_meta_get(
        self, workspace_id: str, experiment_id: str, case_id: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/custom-artifacts"
        ).resolve()
        return self._http_client.get_json(url)

    def custom_artifact_upload(
        self,
        path_to_artifact: str,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        artifact_id: Optional[str] = None,
        overwrite: bool = False,
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/custom-artifacts-imports"
        ).resolve()
        options: Dict[str, Any] = {"overwrite": overwrite}
        if artifact_id:
            options["artifactId"] = artifact_id
        with open(path_to_artifact, "rb") as f:
            multipart_form_data = {
                "file": f,
                "options": json.dumps(options),
            }
            return self._http_client.post_json(url, files=multipart_form_data)

    def case_result_upload(
        self,
        path_to_result: str,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        overwrite: bool = False,
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/experiments/{experiment_id}/cases/"
            f"{case_id}/result-imports"
        ).resolve()
        options: Dict[str, Any] = {"overwrite": overwrite}
        with open(path_to_result, "rb") as f:
            multipart_form_data = {
                "file": f,
                "options": json.dumps(options),
            }
            return self._http_client.post_json(url, files=multipart_form_data)



from __future__ import annotations

import os
import tempfile
from typing import Any, Optional, Text, Union

from modelon.impact.client.entities.file_uri import CustomArtifactURI
from modelon.impact.client.entities.interfaces.custom_artifact import (
    CustomArtifactInterface,
)
from modelon.impact.client.operations.base import BaseOperation
from modelon.impact.client.operations.custom_artifact import (
    CustomArtifactImportOperation,
)
from modelon.impact.client.sal.experiment import ExperimentService


class CustomArtifact(CustomArtifactInterface):
    """CustomArtifact class."""

    def __init__(
        self,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        artifact_id: str,
        download_as: str,
        exp_sal: ExperimentService,
    ):
        self._workspace_id = workspace_id
        self._exp_id = experiment_id
        self._case_id = case_id
        self._artifact_id = artifact_id
        self._download_as = download_as
        self._exp_sal = exp_sal
        super().__init__(
            workspace_id, experiment_id, case_id, artifact_id, download_as, exp_sal
        )

    @property
    def id(self) -> str:
        """Id of the custom artifact."""
        return self._artifact_id

    @property
    def download_as(self) -> str:
        """File name for the downloaded artifact."""
        return self._download_as

    def download(self, path: Optional[str] = None) -> str:
        """Downloads a custom artifact. Returns the local path to the downloaded
        artifact.

        Args:
            path: The local path to the directory to store the downloaded custom
                artifact. Default: None. If no path is given, custom artifact
                will be downloaded in a temporary directory.

        Returns:
            path: Local path to the downloaded custom artifact.

        Example::

            artifact_path = artifact.download()
            artifact_path = artifact.download('/home/Downloads')

        """
        artifact, _ = self._exp_sal.case_artifact_get(
            self._workspace_id, self._exp_id, self._case_id, self.id
        )
        if path is None:
            path = os.path.join(tempfile.gettempdir(), "impact-downloads")
        os.makedirs(path, exist_ok=True)
        artifact_path = os.path.join(path, self.download_as)
        with open(artifact_path, mode="wb") as f:
            f.write(artifact)
        return artifact_path

    def get_data(self) -> Union[Text, bytes]:
        """Returns the custom artifact stream.

        Returns:
            artifact: The artifact byte stream.

        Example::

            artifact = case.get_artifact("ABCD")
            data = artifact.get_data() # may raise exception on communication error
            with open(artifact.download_as, "wb") as f:
                f.write(data)

        """
        result_stream, _ = self._exp_sal.case_artifact_get(
            self._workspace_id, self._exp_id, self._case_id, self.id
        )

        return result_stream

    def get_uri(self) -> CustomArtifactURI:
        """Returns a CustomArtifactURI class.

        Returns:
            The CustomArtifactURI class object.

        Example::

            artifact_file_uri = artifact.get_uri()

        """
        return CustomArtifactURI(self._exp_id, self._case_id, self._artifact_id)

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[CustomArtifact], **kwargs: Any
    ) -> CustomArtifact:
        assert isinstance(operation, CustomArtifactImportOperation)
        return cls(**kwargs, exp_sal=operation._sal.experiment)
